#include "main.h"
/*****************************************************************
ICO tripmaster working as follows:
LCD segments (4 digit standard LCD connected to ports A-D according
to main.h definitions). Buttons on Port E0 (-)  and Port E1 (+), reed
contact towards GND from pin 12 (IRQ0) with  a 1F capacitor fordebouncing and a 
resistor 4K7 to PE2.
Two interrupts work as follows:

External interrupt0 on pin 12:
Configured as input with pullup resistor and capacitor for debouncing
of reed contact that samples the number of wheel spins. Calculates the 
travelled trip every time it fires. Supported by another pullup from PE2
that supplies additional current to contact while device is turned on (4K7).

Timer2 interrupt:
Fires every 15 ms (64Hz)and switches the ports A-D according to the required
data to be displayed on the LCD (32Hz LCD frequency). Reads the two 
buttons every cycle and controls the operation of the system according 
to the buttons pressed and required time delays (debouncing included).

Four display modes:
1. Trip: Displays the currently calculated trip on the LCD. Allows with 
+ or - button to increase or decrease the displayed trip with convenient
button delays.
When trip is on 0km and - button is pressed long enough the device turns off.
Press both buttons to change to mode speed.

2. Speed mode:
In trip mode press both keys long enough. Speed mode will be selected, the 
comma will vanish and the trip symbol will show up. Press the plus key long 
enough to change to max_speed mode. It will show the max speed symbol and the 
maximal speed measured. Reset the max speed with the minus button, go back to
speed mode with both keys.
Press both keys in order to change to clock mode. Press plus key to go to circum mode.

3. Circum: In speed mode press - key long enough.Displays the currently configured 
wheel circumferance.
Operation: The first digit blinks on mode entry.
Then use + button to increase (8-9-0-1-2) the digit to the desired
value or use the - button to change to next digit (1-2-3-4-1).
Press both keys to quit circum mode and to save the value.

4. Clock mode:
In speed mode press both keys.
Displays clock with colon. Use + key to change to set mode (display blinks).
+ key changes selected digit, - key changes to next digit. Both keys change back
to trip mode.
*****************************************************************/

//****************************************************************
//********* global Variables *************************************
//****************************************************************
const unsigned char segments[18]= {0x3F,0x06, 0x5B,0x4F,0x66,0x6D,0x7D,0x07,0x7F,0x6F,0x00,0x40,0x20,0x58,0x01, 0x08}; // segment patterns
volatile unsigned char oldkey, key, mode=trip, Display[4]= {0,0,0,0}, delay1, delay2, delay3, blink, change, count, veloc, old_veloc, max_veloc, power;
volatile unsigned char hour=0, minute=0, four_second=0, num[8], oldTC;
volatile unsigned int circum=2160, trip_100m=0, revolution, temp; 
volatile unsigned long trip_mm=0;					//counts trip in mm up to 100m

unsigned char WaitForKey (unsigned char);
void Switch2Speed(void);

int main(void)
{
	//Fuse CKDIV8 und JTAGEN aus!!
	MCUCR |= (1<<ISC01);		//INT0 falling slope
	GICR |= (1<<INT0);			//INT0 on
	ACSR |= (1 << ACD);		//Turn off analog comparator for power reduction
	/* Enable the Timer/Counter2 interrupts by setting OCIE2 and TOIE2. */
	TIMSK &= ~((1<<OCIE2)|(1<<TOIE2));
	/* Write new values to TCNT2, OCR2, and TCCR2. */ 
	TCNT2 = preload;
	TCCR2 |= ((1<<CS20) |(1<<CS22)); 			//128 prescale Timer Interrupt           
	TCCR2 &= ~(1<<CS21);						//for 64Hz Timer Interrupt
	TIFR |= ((1<<OCF2)|(1<<TOV2));			//clear Timer/Counter2 Interrupt Flags.
	TIMSK|= (1<<TOIE2);						//Timer 2 Overflow Interrupt ein.
	DDRA = 0xFF; 
	DDRB = 0xFF; 
	DDRC = 0xFF; 
	DDRD = 0xFB; 						//All Ports OUTPUTS except INT0 (Pin12)
	DDRE = 0xFC;						//PortE is input on PE0 and PE1
	PORTE = 7;							//Pullups on here also port E             
	max_veloc=0;
	power=1;
	sei();								//Enable all interrupts
	set_sleep_mode(SLEEP_MODE_IDLE);
    while(1)                    
    { 
	 sleep_mode();
    }
	return (1);
}

//****************************************************************
//*************external Interrupt*********************************
//****************************************************************
SIGNAL (SIG_INTERRUPT0)	
{
	//completion in about 2,7ms @ 500KHz CPU clock
	cli ();					//Global Interrupts off
	unsigned char curr_TC = TCNT2;
	unsigned int time;
	unsigned long loc_tmp;
	trip_mm+=circum;		//add wheel circum to mmtrip
	if (trip_mm > 99999)
		{
		 trip_mm -= 100000;	//100m are over	
		 trip_100m++;
		 if (trip_100m > 9999)
			trip_100m=0;
		}
	if ((revolution*steps)> slowest)
		veloc=0;
	else
	{
	   time = curr_TC-oldTC + revolution*steps;
	   loc_tmp= circum;
	   loc_tmp= frequ*36*loc_tmp/time;
	   veloc=loc_tmp/10;
	   if ((veloc==old_veloc)&&(veloc>max_veloc))
			max_veloc=veloc;		//store maximum speed
	   old_veloc=veloc;				//remember last speed
	}
	revolution=0;
	oldTC=curr_TC;
	sei();
//End of wheel interrupt....	
}
//******************************************
//Timer 2 Interrupt here: ******************
//******************************************
SIGNAL(SIG_OVERFLOW2)				//fires about every 15.6ms, 64Hz, takes about 4ms to complete @500KHz CPU clock	
{
	TCNT2 = preload;
	unsigned char i;
	int j;
	cli();							//Global Interrupts off
	blink++;
	revolution++;
	if ((revolution*steps)> slowest)
	  {
		veloc=0;
		revolution--;
	  }	
	i= blink & 0x01;				//can be 0 or 1 toggling every 15ms
	PORTA=num[4*i];				//send data to display
	PORTB=num[4*i+1];	
	PORTC=num[4*i+2];	
	PORTD=num[4*i+3];
	 // Display processing over at this point	
	if ((!blink) && (++four_second ==15))
	{ 
	  four_second=0;
	  if (++minute > 59)
	   { 
	    minute=0;
		if (++hour > 23)
			hour=0;
	   }
	}
	key = (PINE&0x03);
	if (power && !(blink&0x03))		//power saving 16 times a second only
	   {
		for (i=0; i<4; i++)
			num[i]=0;
	PORTD |= (1<<PD2);					//Pullup-resistor on INT0
	DDRE = 0xFC;						//port E2 is output
	PORTE |= (1<<4);					//additional Pullup on
	   }
	if (!power)
	{
	 for (i=0; i<8; i++)
		num[i]=0xFF;		//All segments OFF
	 mode=trip;
	 if (WaitForKey(key))	//Check if wakeup key was presssed
		{
		 delay1 = wait1;
		 delay2 = wait2;
		 delay3 = wait4;
		}
	 PORTD &= ~(1<<PD2);	//Pullup-resistor off INT0
	 DDRE = 0xF8;			//pullup for external contact off, PE2 is input
	 PORTE &= ~(1<<4);
	}
	else
	{
	switch (mode)
	{
	 case shutdown:		//turn display off
		delay1 = wait1;
		delay2 = wait2;
		delay3 = wait4;
		if (--power)
		 {
		  Display[0]=empty;
		  Display[1]=empty;
		  Display[2]=empty;
		  Display[3]=empty;
		 }
	 break;

	// *************Trip processing ***********************
	case trip:
	   if ((change)&&!(blink&0x03))		//16 times a second only, recuduces power consumption
	   {
	     Display[0]=(trip_100m/1000)%10;
		 if (!Display[0])
			Display[0]=empty;
		 Display[1]=(trip_100m/100)%10;
		 if ((Display[0]==empty)&&(!Display[1]))
			Display[1]=empty;
		 Display[2]=(trip_100m/10)%10;
		 Display[3]=trip_100m%10;
	   }
	   if (WaitForKey(key))		//same key pressed long enough...
		 {
		  if (key_min==key)
			{
				if (trip_100m)		//decrease only if trip_100m >0
				   {
					trip_100m--;
					trip_mm=0;
					change = 1;
				   }
				else
				{
				  if (change !=1)	//0: Display Trip off, 2: Display Trip on, no key pressed in between
				  {
				   delay1 = 3*wait1;
				   change = 0;
				   for (i=0; i<4; i++)
				    {
				     if (delay3 < wait4-i)
						 Display[i]=turnoff;
				    }
					 if (delay3==wait4-5)	//wait until - key preesed again on Trip = 0
					 {
					  power=50;			//Turn off here....
					  mode=shutdown;
					 }
						
				   }
				}
			}
			if (key_plus==key)
			 {
			 	trip_mm=0;
				trip_100m++;
				if (trip_100m > 9999)
					trip_100m=0;
			 }
			if (key_all==key)
			 {
				Switch2Speed();			//use both buttons to go to next mode.
				delay1 = 2*wait1;
			 }
		 }
	num[3]|=Komma;				//Komma on
	break;
	//************** Trip over, clock mode ******************************  
    case clock:
	    Display[0]= empty;
		if (hour/10)
			Display[0]=hour/10;
		Display[1]=hour%10;
		Display[2]=minute/10;
		Display[3]=minute%10;
		if (blink & 0x20)
			num[2]|=Colon;				//Colon on when required  
	    if (WaitForKey(key))			//same key pressed long enough...
		{
		  if (key_min==key)
			{
			 mode=setclk;
			 delay1=2*wait1;
			 count=3;					//start with minutes	
			}
		  if (key_all==key)
		   {
			mode=trip;					//use both buttons to go to next mode.
			delay1 = 2*wait1;
		   }
		}
	break;
	case setclk: 
	    Display[0]= empty;
		if (hour/10)
			Display[0]=hour/10;
		Display[1]=hour%10;
		Display[2]=minute/10;
		Display[3]=minute%10;
		if (blink & 0x20)
			num[2]|=Colon;				//Colon on when required  
		 if (blink & 0x10)
		   {
			Display[count] = empty;		//blink  appropriate digits
			Display[count-1] = empty;	
		   }
		 if (WaitForKey(key))
		 { 
		  if (key_all==key)
		   {
			mode=clock;					//use both buttons to go end set  mode.
			delay1 = 2*wait1;
		   }
		  if (key_min==key)
		   {
			 delay1=3*wait1;
			 count=4-count;				//toggle	
		   }
		  if (key_plus==key)
		   {
		    delay1=wait1;
			four_second=0;
			if (count==3)
			   {
				if (++minute>59)
					minute=0;
			   }
			 else
               {
				if (++hour>23)
					hour=0;
			   }	
			}
		 }
		break;
    // *************** clock over *********************************  
	case speed:
	  if (change)
	  {
	   Display[0]= speedmark;
	   if (!(blink&0x3F))
	   {
	    Display[1]=veloc/100;
	    if (!Display[1])
			Display[1]=empty;
	    Display[2]=(veloc/10)%10;
	    if ((Display[1]==empty)&&(!Display[2]))
			Display[2]=empty;
	    Display[3]=veloc%10;
	   }
	  }
	  if (WaitForKey(key))		//same key pressed long enough...
	  {
	   if (key_all==key)
		{
		 mode=clock;			//use both buttons to go to next mode.
		 count=3;
		 delay1 = 3*wait1;
		 }
	   if (key_plus==key)
		{
		 mode=maxspeed;			//use plus button to go to next mode.
		 count=3;
		 delay1 = 3*wait1;
		 } 
	   if (key_min==key)
		{
		 change=0;
		 delay1 = 3*wait1;
		 for (i=0; i<4; i++)
		 {
		  if (delay3 < wait4-i)
			Display[i]=cir;
		 }
		 if (delay3==wait4-5)	//wait until - key preesed again on Trip = 0
			mode = wheel;
			count=1;			//start with second left digit (Display[1])
			temp = circum;
		}
	   }
	break;
	case maxspeed:
	  if (WaitForKey(key))		//same key pressed long enough...
	  {
	   if (key_all==key)
		{
		 Switch2Speed();		//use both buttons to go to next mode.
		 count=3;
		 delay1 = 3*wait1;
		}
	   if (key_min==key)
		{
		 max_veloc=0;
		}
	  }
	  Display[0]= maxspeedmark;
	  Display[1]=max_veloc/100;
	   if (!Display[1])
			Display[1]=empty;
	   Display[2]=(max_veloc/10)%10;
	   if ((Display[1]==empty)&&(!Display[2]))
			Display[2]=empty;
	   Display[3]=max_veloc%10;
	break;
	 // **************** speed over ******************************
	case wheel:
		Display[0]= temp/1000;
		Display[1]= (temp-(Display[0]*1000))/100;
		Display[2]= (temp-Display[0]*1000-Display[1]*100)/10;
		Display[3]= temp-Display[0]*1000-Display[1]*100-Display[2]*10;
		if (blink & 0x10)
			Display[count] = empty;
		if (WaitForKey(key) && (change == 2))		//same key pressed long enough...
		{
		  if (key_min == key)
		   {
		    count++;
			if (count >3)
				count = 1;
			delay1 = wait1;
		   }
		  if (key_plus == key)
		   {
		    delay1 = wait1;
		    switch (count)
			{
			 case 1: temp += 100;
			 if (temp > 3000)
				temp -= 1000;
			 break;
			 case 2: 
			 j=temp/100;
			 temp += 10;
			 if (temp > 100*j+99)
				temp -= 100;
			 break;
			 case 3: 
			 j=temp/10;
			 temp ++;
			 if (temp > 10*j+9)
				temp -= 10;
			}
		   }
		  if (key_all==key)
		   {
		    Switch2Speed();			//use both buttons to go to next mode.
		    circum = temp;
		    delay1 = 2*wait1;
		   }
		}
	break;
	default:
		mode=trip;
	break;
	 } 
	}
	if (power && !(blink&0x03))		//power saving 16 times a second only
	{
	 //First digit segment A:
	 if(segments[Display[0]]&1)	//Filter segment A out
		num[0] |=S1A;
	 //second segment A:
	 if(segments[Display[1]]&1)	//Filter segment A out
		num[0]|=S2A;
	 //third segment A:	
	 if(segments[Display[2]]&1)	//Filter segment A out
		num[2] |=S3A;
	 //fourth segment A:
	 if(segments[Display[3]]&1)	//Filter segment A out
		num[3]|=S4A;
	//*********************************************************	
	 //First digit segment B:
	 if(segments[Display[0]]&2)	//Filter segment B out
		num[0] |=S1B;
	 //second digit segment B:
	 if(segments[Display[1]]&2)	//Filter segment B out
		num[2]|=S2B;
	 //third digit segment B:	
	 if(segments[Display[2]]&2)	//Filter segment B out
		num[2] |=S3B;
	 //fourth digit segment B:
	 if(segments[Display[3]]&2)	//Filter segment B out
		num[3] |=S4B;
	//********************************************************* 
	 //First digit segment C:
	 if(segments[Display[0]]&4)	//Filter segment C out
		num[1] |=S1C;
	 //second digit segment C:
	 if(segments[Display[1]]&4)	//Filter segment C out
		num[1] |=S2C;
	 //third digit segment C:
	 if(segments[Display[2]]&4)	//Filter segment C out
		num[3] |=S3C;
	 //fourth digit segment C:
	 if(segments[Display[3]]&4)	//Filter segment C out
		num[3] |=S4C;
	//********************************************************* 
	 //First digit segment D:
	 if(segments[Display[0]]&8)	//Filter segment D out
		num[1] |=S1D;
	 //second digit segment D:
	 if(segments[Display[1]]&8)	//Filter segment D out
		num[1] |=S2D;
	 //third digit segment D:
	 if(segments[Display[2]]&8)	//Filter segment D out
		num[1] |=S3D;
	 //fourth digit segment D:
	 if(segments[Display[3]]&8)	//Filter segment D out
		num[3]|=S4D;
	//********************************************************* 
	 //First digit segment E:
	 if(segments[Display[0]]&16)	//Filter segment E out
		num[1] |=S1E;
	 //second digit segment E:
	 if(segments[Display[1]]&16)	//Filter segment E out
		num[1] |=S2E;
	 //third digit segment E:
	 if(segments[Display[2]]&16)	//Filter segment E out
		num[1] |=S3E;
	 //fourth digit segment E:
	 if(segments[Display[3]]&16)	//Filter segment E out
		num[3] |=S4E;
	 //********************************************************* 
	 //First digit segment f:
	 if(segments[Display[0]]&32)	//Filter segment f out
		num[0] |=S1F;
	 //second digit segment f:
	 if(segments[Display[1]]&32)	//Filter segment f out
		num[0] |=S2F;
	 //third digit segment f:
	 if(segments[Display[2]]&32)	//Filter segment f out
		num[2] |=S3F;
	 //fourth digit segment f:
	 if(segments[Display[3]]&32)	//Filter segment f out
		num[2] |=S4F;
	 
	 //********************************************************* 
	 //First digit segment G:
	 if(segments[Display[0]]&64)	//Filter segment G out
		num[0] |=S1G;
	 //second digit segment G:
	 if(segments[Display[1]]&64)	//Filter segment G out
		num[0] |=S2G;
	 //third digit segment G:
	 if(segments[Display[2]]&64)	//Filter segment G out
		num[2] |=S3G;
	 //fourth digit segment G:
	 if(segments[Display[3]]&64)	//Filter segment G out
		num[2] |=S4G;
		
	 for (i=0; i<4; i++)
		num[i+4] = num[i] ^ 0xFF;
	 num[3]|=0x04;					//keep pullup on
	 num[7]|=0x04;							
	}
	oldkey = key;
	sei();    
}

// **********************************************************
// checks if keys were pressed long enough and controls delays
// Return: 1, take key action or 0, take no action
// **********************************************************
unsigned char WaitForKey (unsigned char KeyPressed)
{
  if ((KeyPressed != key_none)&&(oldkey == KeyPressed))
	{
	 if (!--delay1)		//wait for key to be pressed long enough before action
	 {
	  delay1++;				//same key pressed long enough
	  if (!--delay2)		//loop slowly here
	  {
	    delay2=wait2;		//loop slowly
		power=1;
		if (!--delay3)
		{
			delay3++;
			delay2=wait3;	//now loop faster
		}
		return (1);
	  }
	 }
	}
	else
	{
	 delay1 = wait1;
	 delay2 = wait2;
	 delay3 = wait4;
	 if((!change)&&(mode==speed))
		Switch2Speed();
	 change = 2;			//Trip display enabled, flag that keys were released last
	}
  return (0);
}

// **********************************************************
// switch to speed mode with no display delay
// **********************************************************
void Switch2Speed(void)	
{
 mode=speed;
 Display[0]= speedmark;
 Display[1]=veloc/100;
 if (!Display[1])
    Display[1]=empty;
 Display[2]=(veloc/10)%10;
 if ((Display[1]==empty)&&(!Display[2]))
    Display[2]=empty;
 Display[3]=veloc%10;
}
